/**
* =====================================================================================
*       struct: tmp  
*  Description: temporary structure - initialization of the tree 
* =====================================================================================
*/
struct tmp
  {
    double      time;
    double      prop;
    reaction*   r;
    tmp(double,double,reaction* p=NULL);
  };
tmp::tmp(double init_time,double init_prop,reaction* p)
    :time(init_time),prop(init_prop),r(p)
      {}
 /**
 * =====================================================================================
 *  struct: reac_info   
 *  Description: update and manipulation of delayed reactions 
 * =====================================================================================
 */
struct reac_info
 {
    double     time;                        ///<reaction time 
    double     prop;                        ///<propensity of the reaction
    reaction*  r;                           ///<index of the reacion
    double     time_last_nonzero;           //
    double     time_last_nonzero_reaction;  //case of prop=0 (time=inf)
    double     prop_last_nonzero;           //redundant random number 

    priority_queue<double,vector<double>,greater<double> > delayed_times_que;   

    void put(double i_time,double i_prop, reaction* i_r);
    reac_info (double init_time=INFINITY, double init_prop=0.0, reaction* init_r=NULL);
    void put_into_que(double delayed_time);
    void take_top_from_que();
    void ascribe_time();
    void clear();
 };
reac_info::reac_info(double init_time, double init_prop, reaction* init_r)
      :time(init_time),prop(init_prop),r(init_r),
       time_last_nonzero(0.0),time_last_nonzero_reaction(0.0),prop_last_nonzero(0.0)
    {}  
void reac_info::put(double i_time,double i_prop, reaction* i_r)
  {
    time=i_time;
    prop=i_prop;
    r=i_r;
  }
void reac_info::clear()
  {
    while(!delayed_times_que.empty())
    {
      delayed_times_que.pop();
    }
  }
void reac_info::put_into_que(double delayed_time)
  {
    delayed_times_que.push(delayed_time);
  }
void reac_info::take_top_from_que()
  {
    if(!delayed_times_que.empty()) 
    {
    delayed_times_que.pop();
    }
  }
void reac_info::ascribe_time()
  {
    if(delayed_times_que.empty())
    {
      time=INFINITY;    
    }
    else
    {
      time=delayed_times_que.top();
    }
  }
  /**
 * =====================================================================================
 *  Function object: comp   
 *  Description: greater for tmp.time
 * =====================================================================================
 */
struct comp
{
  bool operator()(reac_info t1, reac_info t2) const
    {
      return (t1.time<t2.time);
    }  
};
/**
 * =====================================================================================
 *        struct: node   
 *  Description: contains informations for construction of a tree 
 * =====================================================================================
 */
 struct node
  {
      reac_info* ri;                   ///<pointer to reaction info
      node*      left_child;           ///<pointer to the left node (lesser)
      node*      right_child;          ///<pointer to the right node (greater)
      node*      parent;               ///<pointer to the parent node

      node (node* init_left=NULL,node* init_right=NULL,reac_info* init_ri=NULL);  
  };
node::node(node* init_left,node* init_right,reac_info* init_ri)
      :left_child(init_left),right_child(init_right),ri(init_ri)
    {}  
/**
 * =====================================================================================
 *        Class:  NEXT REACTION METHOD \n
 *  Description:  structure for next reaction method \n
 *  Method for simulating time evolution of reactions:
 *  M. A. Gibson and J. Bruck, J. Phys. Chem. A 104,1876(2000)
 * =====================================================================================
 */
class NEXT_REACTION_METHOD
{
  private:
  root&  r;                                  ///<contains connections between reagents and reactions 
  gsl_rng*  rand_gen;                        ///<generator of random numbers
  public:
  vector<reaction*> vect_p_reactions;        ///<vector of pointers to the reactions calculated using G_D_M
  vector<node*> index_structure;             ///<index structure
  size_t n;                                  ///<number of reactions calculated using this method
  node* tree;                                ///<pointer to the root of the tree
  reac_info* tab_reac_info;                  ///<vector of structures reac_inf
  int levels;                                ///<number of levels in the tree structure (root=0)
  reac_info* helper;                         ///<helper for swap()
  
  double new_propensity;                     ///<a new propensity
  reaction* root_reac;                       ///<reaction in the root
  reac_info* curr_ri;                        ///<reac_info currently changed in the node; 
  //node* min_node;                            ///<pointer to node with minimal reaction time 
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 
  NEXT_REACTION_METHOD(root& init_root,gsl_rng* init_rand);  /* constructor */
  ~NEXT_REACTION_METHOD();
  void reinitialize();
  int pow2(int i);
  int get_levels(); 
  double calculate_propensity(reaction* react);
  inline double calculate_time(const double prop);
  inline double calculate_time_delayed(const double reaction_rate_constant, const double number_of_elementary_reactions);
  void swap(node* n,node* m);  
  void update(node* n);
  void react();  
  void check_reaction(double&,reaction*&) ;
  void recalculate_and_update(node* n);
  void recalculate_and_update_root_after_jump(node* n,double previous_root_time);  
  void conc_jump(int mol_n,int mol_i,double curr_time);
  void division(double curr_time);
  void rrc_jump(double rrc_const,int r_ind,double curr_time);
  void diff_react_update(size_t _r);
  void recalculate_react_rate_const();
};  
NEXT_REACTION_METHOD::NEXT_REACTION_METHOD (root& init_root,gsl_rng* init_rand)
                    :r(init_root),rand_gen(init_rand),levels(0),tree(NULL),
                     index_structure(r.vect_reactions.size(),NULL)
  {
    //fullfil vector of pointers
    for(size_t i(0),j=r.vect_reactions.size();i<j;++i)
    {  
       //put here only "Gillespie's" reactions
       if(r.vect_reactions[i].if_g || r.vect_reactions[i].if_pure_g)
       {
         vect_p_reactions.push_back( &(r.vect_reactions[i]) ); 
       }
    }
    //number of reactions
    n=vect_p_reactions.size();    
    //ascribe appropriate reaction rate constants
    recalculate_react_rate_const();
    //make a binary tree
    tree=new node [n];
    tab_reac_info=new reac_info [n];
    get_levels();
    for(int a(0);a<=levels;++a)
    { 
      for(int b(0),c=pow2(a),d=pow2((a+1));b<c;++b)
      { 
        //offspring
        if((c-1)+b>=n) continue;   //eliminate not existed 
        if(((d-1)+2*b)<n)
        tree[(c-1)+b].left_child=&tree[(d-1)+2*b];
        else {tree[(c-1)+b].left_child=NULL;}
        if(((d-1)+(2*b+1))<n)
        tree[(c-1)+b].right_child=&tree[(d-1)+(2*b+1)];
        else {tree[(c-1)+b].right_child=NULL;}
        //parent
        if(a!=0) {tree[(c-1)+b].parent=&tree[(pow2((a-1))-1)+(b/2)];}
        else     {tree[(c-1)+b].parent=NULL;}
      }
    }
    //calculate propensities and times for reactions
    double p;
    for(size_t i(0);i<n;++i)
      {
        //delayed reactions (one substrate)->products
        if(vect_p_reactions[i]->delayed)
        {
          for(size_t j(0);j<vect_p_reactions[i]->n_subs;++j)
          {
            for(size_t k(0);k<(r.vect_molecules[ vect_p_reactions[i]->subs_ind[j] ])->mol_numb;++k)
            {
               tab_reac_info[i].put_into_que(calculate_time_delayed(vect_p_reactions[i]->react_rate_const,vect_p_reactions[i]->number_of_steps));
            }
          } 
          //ascribe proper time from delayed reaction in the priority que
          tab_reac_info[i].ascribe_time();
          tab_reac_info[i].r=vect_p_reactions[i];
          continue;
        }
        p=calculate_propensity(vect_p_reactions[i]);
        if(p==0.0)
        {
          tab_reac_info[i].put(INFINITY,
                               p,
                               vect_p_reactions[i]);
        }
        else
        {
          tab_reac_info[i].put(calculate_time(p),
                               p,
                               vect_p_reactions[i]);
        }
      }
    int NN = sizeof(*tab_reac_info)*n/sizeof(reac_info);
    sort(tab_reac_info,tab_reac_info+NN,comp());      
    for(size_t i(0);i<n;++i)
      { 
         tree[i].ri=&tab_reac_info[i];
      } 
    //make index structure
    for(size_t i(0);i<n;++i)
      {
        index_structure[((tree[i].ri)->r)->reac_ind]=&tree[i];               
      }
  }
NEXT_REACTION_METHOD::~NEXT_REACTION_METHOD()
  {
    delete []tree;
    delete []tab_reac_info;
  }
void NEXT_REACTION_METHOD::reinitialize()
{
double p;
    //clear delayed
    for(size_t i(0);i<n;++i)
    {
      if(tab_reac_info[i].r->delayed)
      {
        tab_reac_info[i].clear();
      }
    }
    for(size_t i(0);i<n;++i)
      {
       //delayed reactions(one substrate)->products
        if(vect_p_reactions[i]->delayed)
        {
          for(size_t j(0);j<vect_p_reactions[i]->n_subs;++j)
          {
            for(size_t k(0);k<(r.vect_molecules[ vect_p_reactions[i]->subs_ind[j] ])->mol_numb;++k)
            {
               tab_reac_info[i].put_into_que(calculate_time_delayed(vect_p_reactions[i]->react_rate_const,vect_p_reactions[i]->number_of_steps));
            }
          } 
          //ascribe proper time from delayed reaction in the priority que
          tab_reac_info[i].ascribe_time();
          tab_reac_info[i].r=vect_p_reactions[i];
          continue;
        } 
        p=calculate_propensity(vect_p_reactions[i]);
        tab_reac_info[i].put(calculate_time(p),
                             p,
                             vect_p_reactions[i]);
      }
    int NN = sizeof(*tab_reac_info)*n/sizeof(reac_info);
    sort(tab_reac_info,tab_reac_info+NN,comp());      
    for(size_t i(0);i<n;++i)
      { 
         tree[i].ri=&tab_reac_info[i];
         (tree[i].ri)->time_last_nonzero=0.0;
         (tree[i].ri)->time_last_nonzero_reaction=0.0;
         (tree[i].ri)->prop_last_nonzero=0.0;
      } 
    //make index structure
    for(size_t i(0);i<n;++i)
      {
        index_structure[((tree[i].ri)->r)->reac_ind]=&tree[i];               
      }
}
int NEXT_REACTION_METHOD::pow2(int i)
  {
    if(i==0) return 1;
    return pow2(i-1)*2;
  } 
int NEXT_REACTION_METHOD::get_levels() 
  {
    int _n=n;
    for(;;)
    {
      _n-=pow2(levels);
      if(_n<=0) break;
      ++levels;
    }
  }
double NEXT_REACTION_METHOD::calculate_propensity(reaction* react)
  {
    double propensity=react->react_rate_const;
    if(react->no_simple)
    {  
       size_t index=0;
       int number_same=1;  
       for(size_t i(0);i<react->n_subs;++i)
      {
        if(i==0)
        {
          index=react->subs_ind[i];
          propensity*=(r.vect_molecules[index])->mol_numb; 
        }
        else
        {  
          if(index == react->subs_ind[i])
            {
              ++number_same;
              propensity*=(1.0/(double)number_same)*
                          ((r.vect_molecules[index])->mol_numb-((double)number_same-1.0)); 
            }
          else
            {
              index=react->subs_ind[i];
              propensity*=(r.vect_molecules[index])->mol_numb;
              number_same=1;
            }
        } 
      }
      //no reagents
      if(propensity<=0.0 ) propensity=0.0;     
    }
    else
    {
      for(size_t i(0);i<react->n_subs;++i)
      {
        propensity*=(r.vect_molecules[ react->subs_ind[i] ])->mol_numb;   
      }
    }  
    return propensity;    
  }
inline double NEXT_REACTION_METHOD::calculate_time(const double prop)
  {
    return gsl_ran_exponential(rand_gen, 1.0/prop);
  }
inline double NEXT_REACTION_METHOD::calculate_time_delayed(const double reaction_rate_constant, const double number_of_elementary_reactions)
  {
    return gsl_ran_gamma(rand_gen, number_of_elementary_reactions, 1.0/reaction_rate_constant);
  }
void NEXT_REACTION_METHOD::swap(node* n,node* m)  
  {
      //swap reaction_info
      helper=n->ri;n->ri=m->ri;m->ri=helper;
      //update index_structure
      index_structure[((n->ri)->r)->reac_ind]=n;
      index_structure[((m->ri)->r)->reac_ind]=m;
  }
void NEXT_REACTION_METHOD::update(node* n)
  {
 node* min_node;   if(n->parent != NULL && (n->ri)->time < ((n->parent)->ri)->time)
      {
        swap(n,(n->parent));
        update((n->parent));
      } 
    else 
       {
         if(n->right_child == NULL)  min_node=n->left_child;
         else
          {
            if(((n->left_child)->ri)->time > ((n->right_child)->ri)->time) 
            min_node=n->right_child;
            else                                                           
            min_node=n->left_child;
          }
         if(min_node == NULL) return;
         if((n->ri)->time > (min_node->ri)->time)
          {
            swap(n,min_node);
            update(min_node);
          }
         else return;
       }
  }
void NEXT_REACTION_METHOD::recalculate_and_update(node* n)
  {
    curr_ri=n->ri;
    if((curr_ri->r)->delayed)
    {
      curr_ri->put_into_que((tree[0].ri)->time+
                            calculate_time_delayed((curr_ri->r)->react_rate_const,(curr_ri->r)->number_of_steps));
      curr_ri->ascribe_time();
      update(n);
      return;
    }
    new_propensity=calculate_propensity(curr_ri->r);
    if(curr_ri->time == INFINITY)
      {
        if(new_propensity == 0.0) {return;}
        else
          {
            if((curr_ri->prop_last_nonzero)==0.0) //new random number required (initialization and root)
            {
              curr_ri->prop=new_propensity;
              curr_ri->time=(tree[0].ri)->time+calculate_time(new_propensity);
              update(n);
            }
            else
            {
              curr_ri->time=( (curr_ri->prop_last_nonzero)/(new_propensity)*
                            (curr_ri->time_last_nonzero_reaction-curr_ri->time_last_nonzero)+(tree[0].ri)->time );
              curr_ri->prop=new_propensity;
              update(n);
            }
          }        
      }
    else
      {
        if(new_propensity==0.0)
        {
          curr_ri->time_last_nonzero=(tree[0].ri)->time;
          curr_ri->time_last_nonzero_reaction=curr_ri->time;
          curr_ri->prop_last_nonzero=curr_ri->prop;
          curr_ri->time=INFINITY;
          curr_ri->prop=new_propensity;  
          update(n);
        }
        else
        {
         curr_ri->time=( (curr_ri->prop)/(new_propensity)*(curr_ri->time-(tree[0].ri)->time)+(tree[0].ri)->time );
         curr_ri->prop=new_propensity;  
         update(n);
        }              
      }
  }
void NEXT_REACTION_METHOD::react()  
  {
    root_reac=(tree[0].ri)->r; 
    //go to reactions connected to the edge of the graph
    //for substrates    
    for(size_t i(0);i<root_reac->n_subs;++i)
      {
        //go to substrates to check reactions and update for substrates
        for(size_t j(0) ; j< r.vect_reagents[ root_reac->subs_ind[i] ].n_substrates; ++j)
          {
            //do not update root yet
            if(root_reac->reac_ind != r.vect_reagents[ root_reac->subs_ind[i] ].as_substrate[j])
            {
            recalculate_and_update( 
            index_structure[  r.vect_reagents[ root_reac->subs_ind[i] ].as_substrate[j] ]
                                  ); 
            }
          }  
      }      
    //check products
    for(size_t i(0);i<root_reac->n_prods;++i)  
      {
        //go to products to check reactions and update for substrates
        for(size_t j(0) ; j< r.vect_reagents[ root_reac->prod_ind[i] ].n_substrates; ++j)
          {
            //do not update root yet
            if(root_reac->reac_ind != r.vect_reagents[ root_reac->prod_ind[i] ].as_substrate[j])
            {
            recalculate_and_update( 
            index_structure[  r.vect_reagents[ root_reac->prod_ind[i] ].as_substrate[j] ]
                                           );     
            }
          } 
      }
    //as a last root (chosen reaction has the less time, so it's always in the root)
    if(root_reac->delayed)
    {
     (tree[0].ri)->take_top_from_que();
     (tree[0].ri)->ascribe_time();
     update(tree); 
     return;
    }  
    new_propensity=calculate_propensity(root_reac);
    if(new_propensity==0.0)
    {
      (tree[0].ri)->prop_last_nonzero=0.0; //new random is required if a new propensity !=0.0
      (tree[0].ri)->time=INFINITY;
      (tree[0].ri)->prop=new_propensity;  
    }
    else
    {
      (tree[0].ri)->prop=new_propensity;
      (tree[0].ri)->time+=calculate_time(new_propensity);
    }
    update(tree);
  }
void NEXT_REACTION_METHOD::recalculate_and_update_root_after_jump(node* n,double previous_root_time)
  {
    curr_ri=n->ri;
    new_propensity=calculate_propensity(curr_ri->r);
    if(new_propensity==0.0)
    {
       curr_ri->time_last_nonzero=(tree[0].ri)->time;
       curr_ri->time_last_nonzero_reaction=curr_ri->time;
       curr_ri->prop_last_nonzero=curr_ri->prop;
       curr_ri->time=INFINITY;
       curr_ri->prop=new_propensity;  
       update(n);
    }
    else
    {
       curr_ri->time=( (curr_ri->prop)/(new_propensity)*(previous_root_time-curr_ri->time)+curr_ri->time);
       curr_ri->prop=new_propensity;  
       update(n);
    }
  }
void NEXT_REACTION_METHOD::conc_jump(int mol_n,int mol_i,double curr_time)  
  {
    root_reac=(tree[0].ri)->r; 
    double root_time=(tree[0].ri)->time;
    (tree[0].ri)->time=curr_time;
    (r.vect_molecules[mol_i])->mol_numb=mol_n;
    //go to substrates to check reactions and update for substrates
    for(size_t j(0) ; j< r.vect_reagents[mol_i].n_substrates; ++j)
    {
       if(root_reac->reac_ind != r.vect_reagents[mol_i].as_substrate[j])
       {
         recalculate_and_update( 
                                 index_structure[  r.vect_reagents[mol_i].as_substrate[j] ]
                               );    
       }
    }  
    //as a last root (chosen reaction has the less time, so it's always in the root)
    if(root_reac->delayed)
    {
     (tree[0].ri)->time=root_time;
     update(tree); 
     return;
    }  
    bool appeared=false;
    for(size_t j(0) ; j< r.vect_reagents[mol_i].n_substrates; ++j)
    {
       if(root_reac->reac_ind == r.vect_reagents[mol_i].as_substrate[j])
       {
        appeared=true; 
        recalculate_and_update_root_after_jump( index_structure[r.vect_reagents[mol_i].as_substrate[j]],root_time);      
       }
    }
    if(!appeared)
    {
     (tree[0].ri)->time=root_time;
     update(tree); 
    }  
  }
void NEXT_REACTION_METHOD::division(double curr_time)
  {
     for(size_t j(0) ; j< r.vect_molecules.size(); ++j)
     {
       conc_jump(gsl_ran_binomial(rand_gen,0.5,(r.vect_molecules[j])->mol_numb),j,curr_time);
     }
  }
void NEXT_REACTION_METHOD::rrc_jump(double rrc_const,int r_ind,double curr_time)
{
    root_reac=(tree[0].ri)->r; 
    double root_time=(tree[0].ri)->time;
    (tree[0].ri)->time=curr_time;
    (r.vect_reactions[r_ind]).react_rate_const=rrc_const;
    //go to substrates to check reactions and update for substrates
    if(root_reac->reac_ind != r_ind && !(r.vect_reactions[r_ind]).delayed)
    {
        recalculate_and_update( index_structure[r_ind]);  
    }
    //as a last root (chosen reaction has the less time, so it's always in the root)
    if(root_reac->delayed)
    {
     (tree[0].ri)->time=root_time;
     update(tree); 
     return;
    }  
    if(root_reac->reac_ind == r_ind)
    {
       recalculate_and_update_root_after_jump( index_structure[r_ind],root_time); 
    }
}
//update reaction with diffusive particles 
//an argument is current reaction index
void NEXT_REACTION_METHOD::diff_react_update(size_t _r)
  {
    //for substrates
    reaction& _react=r.vect_reactions[_r];
    node*     tmp;
    for(size_t i(0);_react.n_subs;++i)
      {
        //go to substrates to check reactions and update for substrates
        for(size_t j(0) ; j< r.vect_reagents[ _react.subs_ind[i] ].n_substrates; ++j)
          {
            tmp=index_structure[ r.vect_reagents[ _react.subs_ind[i] ].as_substrate[j] ];
            if(tmp!=NULL) recalculate_and_update(tmp);     
          }  
      }      
    //check products
    for(size_t i(0);_react.n_prods;++i)  
      {
        //go to products to check reactions and update for substrates
        for(size_t j(0) ; j< r.vect_reagents[ _react.prod_ind[i] ].n_substrates; ++j)
          {
            tmp=index_structure[  r.vect_reagents[ _react.prod_ind[i] ].as_substrate[j] ] ;
            if(tmp!=NULL) recalculate_and_update(tmp);     
          } 
      }
  }
//check reaction, if this reaction appears, call react 
void NEXT_REACTION_METHOD::check_reaction(double& _t,reaction*& _r) 
  {
    //first reaction from the tree is current reaction
    _t=(tree[0].ri)->time;
    _r=(tree[0].ri)->r;
  }
//calculate current react_rate_const
void NEXT_REACTION_METHOD::recalculate_react_rate_const() 
  {
    for(size_t i(0);i<n;++i)
      {
       if((vect_p_reactions[i]->n_subs)>1)
        {
          vect_p_reactions[i]->react_rate_const=vect_p_reactions[i]->react_rate_const
                                                /gsl_pow_int(((r.reactor)->get_volume()),
                                                (vect_p_reactions[i]->n_subs)-1);
        }
      }
  }
